from typing import Any

__all__ = [
    "alias_convert",
    "calculate_total_stats",
    "calculate_total_xp",
    "calculate_required_xp",
]


def alias_convert(data: dict, mode: str, model: Any = None):
    if mode == "PLAYER":
        return {PLAYER_ALIASES[i]: data[i] for i in data if i in PLAYER_ALIASES}

    elif mode == "GUILD":
        return {GUILD_ALIASES[i]: data[i] for i in data if i in GUILD_ALIASES}

    elif mode == "GUILD_MEMBER":
        return {
            GUILD_MEMBER_ALIASES[i]: data[i] for i in data if i in GUILD_MEMBER_ALIASES
        }

    elif mode == "LEADERBOARD":
        return {
            LEADERBOARD_ALIASES[i]: data[i] for i in data if i in LEADERBOARD_ALIASES
        }

    elif mode == "ACHIEVEMENT":
        return {
            ACHIEVEMENTS_ALIASES[i]: data[i] for i in data if i in ACHIEVEMENTS_ALIASES
        }

    elif type(GAME_ALIASES[mode]) == dict:
        return {
            i: (
                model.__dataclass_fields__[i].type(**data[GAME_ALIASES[mode][i]])
                if GAME_ALIASES[mode][i] in data
                else model.__dataclass_fields__[i].type(
                    **GAME_STATS[GAME_ALIASES[mode][i]]
                )
            )
            for i in GAME_ALIASES[mode]
        }

    else:
        if GAME_ALIASES[mode] in data:
            return data[GAME_ALIASES[mode]]

        else:
            return GAME_STATS[GAME_ALIASES[mode]]


def calculate_total_stats(data: dict):
    wins = 0
    finals = 0
    kills = 0
    beds = 0
    for i in data:
        try:
            wins += data[i]["wins"]

        except:
            pass

        try:
            finals += data[i]["finals"]

        except:
            pass

        try:
            kills += data[i]["kills"]

        except:
            pass

        try:
            beds += data[i]["beds"]

        except:
            pass

    return {"wins": wins, "finals": finals, "kills": kills, "beds": beds}


def calculate_total_xp(level: int, xp: float):
    if level < 5:
        return level * (level + 1) * 500 - 1000 + xp

    totalxp = 0

    if level < 100 and level >= 5:
        totalxp = 5000 * (level - 1) - 6000 + xp

    if level >= 100:
        sum = 0
        for p in range((level - 100) // 100):
            sum += (((level - 100 * (p + 1)) // 100) * 500 + 5000) * (
                95 - ((level - 100 * (p + 1)) // 200)
            ) + (
                (5 + ((level - 100 * (p + 1)) // 200))
                * (6 + ((level - 100 * (p + 1)) // 200))
                * 500
            )

        if (level - ((level // 100) * 100)) <= (4 + (round((level // 100) / 2))):
            totalxp = (
                (
                    (level - ((level // 100) * 100))
                    * (level + 1 - ((level // 100) * 100))
                    * 500
                )
                + 489000
                + xp
                + sum
            )

        if (level - ((level // 100) * 100)) > (4 + (round((level // 100) / 2))):
            totalxp = (
                ((level // 100) * 500 + 5000)
                * ((level - ((level // 100) * 100)) - (5 + (((level - 100) // 200))))
                + (((5 + ((level - 100) // 200)) * (6 + ((level - 100) // 200))) * 500)
                + 489000
                + xp
                + sum
            )

    return totalxp


def calculate_required_xp(level: int):
    requiredExp = (level // 100) * 500 + 5000
    onesplayer1level = level - (level // 100 * 100)
    levelXP = onesplayer1level * 1000 + 1000
    if requiredExp > levelXP:
        requiredExp = levelXP

    return requiredExp


PLAYER_ALIASES = {
    "uuid": "uuid",
    "lastLoginName": "name",
    "lastLoginTime": "last_login_time_raw",
    "role": "rank",
    "level": "level",
    "exp": "xp",
    "weightedwins": "weightedwins",
    "guildId": "guild_id",
    "joinTime": "join_date_raw",
    "guildRole": "role",
}

GAME_ALIASES = {
    "bedRush": {"single": "bedRushSingle", "double": "bedRushDouble"},
    "bedwalls": "bedwalls",
    "bedwarsLate": {"single": "bedwarsLateSingle", "double": "bedwarsLateDouble"},
    "bedwarsMega": {"single": "bedwarsMegaSolo", "double": "bedwarsMegaDouble"},
    "bedwarsNormal": {"single": "bedwarsNormalSingle", "double": "bedwarsNormalDouble"},
    "bedwarsRush": {"single": "bedwarsRushSolo", "double": "bedwarsRushDouble"},
    "bowFight": {"single": "bowFightSingle", "double": "bowFightDouble"},
    "bridgeFight": {
        "single": "bridgesSingle",
        "double": "bridgesDouble",
        "comp": "compBridgeSingle",
    },
    "flatFight": {"single": "flatFightSingle", "double": "flatFightDouble"},
    "fourWayBridge": "fourWayBridgeSingle",
    "groundFight": {"single": "groundSingle", "double": "groundDouble"},
    "ladderFight": {"single": "ladderFightSingle", "double": "ladderFightDouble"},
    "miniwars": {"single": "miniwarsSolo", "double": "miniwarsDouble"},
    "obstacles": "obstacleSingle",
    "partyGames": "partyGames",
    "pearlFight": {"single": "pearlFightSingle", "double": "pearlFightDouble"},
    "rankedFoursPractice": "rankedFoursPractice",
    "resourceCollect": {"single": "resourceSingle", "double": "resourceDouble"},
    "resourceOldCollect": {
        "single": "resourceOldSingle",
        "double": "resourceOldDouble",
    },
    "stickFight": {"single": "stickFightSingle", "double": "stickFightDouble"},
    "sumo": {"block_sumo": "sumo", "beta_sumo": "betaSumo"},
    "sumoDuels": {"single": "sumoDuelsSolo", "double": "sumoDuelsDouble"},
    "voidFight": {"single": "voidSingle", "double": "voidDouble"},
}

LEADERBOARD_ALIASES = {
    "uuid": "uuid",
    "position": "position",
    "level": "level",
    "weightedwins": "weightedwins",
    "time": "technique_time",
    "submittime": "date_submitted_raw",
    "value": "wins",
}

GUILD_ALIASES = {
    "id": "guild_id",
    "name": "name",
    "desc": "description",
    "xp": "gxp",
    "num": "member_count",
    "ownerUUID": "org_owner",
    "time": "creation_date_raw",
    "placing": "position",
    "tag": "tag",
}

GUILD_MEMBER_ALIASES = {
    "uuid": "uuid",
    "guildId": "guild_id",
    "guildRole": "role",
    "role": "role",
    "joinTime": "join_date_raw",
    "time": "join_date_raw",
}

GAME_STATS = {
    "bedRushDouble": {"wins": 0, "kills": 0},
    "bedRushSingle": {"wins": 0, "kills": 0},
    "bedwalls": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsLateDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsLateSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsMegaDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsMegaSolo": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsNormalDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsNormalSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsRushDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bedwarsRushSolo": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "betaSumo": {"wins": 0, "kills": 0},
    "bowFightDouble": {"wins": 0, "kills": 0},
    "bowFightSingle": {"wins": 0, "kills": 0},
    "bridgesDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "bridgesSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "compBridgeSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "flatFightDouble": {"wins": 0, "kills": 0},
    "flatFightSingle": {"wins": 0, "kills": 0},
    "fourWayBridgeSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "groundDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "groundSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "ladderFightDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "ladderFightSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "miniwarsDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "miniwarsSolo": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "obstacleSingle": {"wins": 0},
    "partyGames": {"wins": 0},
    "pearlFightDouble": {"wins": 0, "kills": 0},
    "pearlFightSingle": {"wins": 0, "kills": 0},
    "rankedFoursPractice": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "resourceDouble": {"wins": 0, "kills": 0},
    "resourceSingle": {"wins": 0, "kills": 0},
    "resourceOldDouble": {"wins": 0, "kills": 0},
    "resourceOldSingle": {"wins": 0, "kills": 0},
    "stickFightDouble": {"wins": 0, "kills": 0},
    "stickFightSingle": {"wins": 0, "kills": 0},
    "sumoDuelsDouble": {"wins": 0, "kills": 0},
    "sumoDuelsSolo": {"wins": 0, "kills": 0},
    "sumo": {"wins": 0, "kills": 0},
    "voidDouble": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
    "voidSingle": {"wins": 0, "finals": 0, "kills": 0, "beds": 0},
}

ACHIEVEMENTS_ALIASES = {
    "id": "achievement_id",
    "name": "name",
    "desc": "description",
    "xp": "xp",
}



from .bed_rush import *
from .bedwalls import *
from .bedwars_late import *
from .bedwars_mega import *
from .bedwars_normal import *
from .bedwars_rush import *
from .bow_fight import *
from .bridge_fight import *
from .flat_fight import *
from .four_way_bridges import *
from .ground_fight import *
from .ladder_fight import *
from .mini_wars import *
from .obstacles import *
from .party_games import *
from .pearl_fight import *
from .ranked_fours_practice import *
from .resource_collect import *
from .resource_old_collect import *
from .stick_fight import *
from .sumo_duels import *
from .sumo import *
from .void_fight import *



from dataclasses import dataclass, field, InitVar
from typing import List, Dict

from ..constants import *
from .player import (
    StatsLeaderboardPlayer,
    TechniqueLeaderboardPlayer,
    PeriodicLeaderboardPlayer,
)
from .guild import LeaderboardGuild

__all__ = [
    "StatsLeaderboard",
    "LevelLeaderboard",
    "WeightedWinsLeaderboard",
    "TechniqueLeaderboard",
    "PeriodicLeaderboard",
    "GuildLeaderboard",
]


@dataclass
class StatsLeaderboard:
    _data: InitVar[Dict] = None
    players: List[StatsLeaderboardPlayer] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        for i in _data:
            self.players.append(StatsLeaderboardPlayer(i))


@dataclass
class LevelLeaderboard(StatsLeaderboard):
    leaderboard: str = "level"


@dataclass
class WeightedWinsLeaderboard(StatsLeaderboard):
    leaderboard: str = "weightedwins"


@dataclass
class TechniqueLeaderboard:
    _data: InitVar[Dict] = None
    players: List[StatsLeaderboardPlayer] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        for i in _data:
            self.players.append(TechniqueLeaderboardPlayer(i))


@dataclass
class PeriodicLeaderboard:
    _data: InitVar[Dict] = None
    game: str = None
    type: str = None
    period: str = None
    players: List[PeriodicLeaderboardPlayer] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        self.game = VOXYL_GAME_NAMES_LOWER[self.game.lower()]

        for i in range(len(_data)):
            self.players.append(PeriodicLeaderboardPlayer(_data[i], position=i + 1))


@dataclass
class GuildLeaderboard:
    _data: InitVar[Dict] = None
    guilds: List[LeaderboardGuild] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        for i in _data:
            self.guilds.append(LeaderboardGuild(i))



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["Sumo", "BlockSumo", "BetaSumo"]


@dataclass
class BlockSumo(WinsKills):
    pass


@dataclass
class BetaSumo(WinsKills):
    pass


@dataclass
class Sumo(WinsKills):
    block_sumo: BlockSumo = None
    beta_sumo: BetaSumo = None

    def __post_init__(self):
        self.wins = self.block_sumo.wins + self.beta_sumo.wins
        self.kills = self.block_sumo.kills + self.beta_sumo.kills



from dataclasses import dataclass
from .default_game_structure import WinsFinalsKillsBeds

__all__ = ["BedwarsLate", "BedwarsLateSingle", "BedwarsLateDouble"]


@dataclass
class BedwarsLateSingle(WinsFinalsKillsBeds):
    pass


@dataclass
class BedwarsLateDouble(WinsFinalsKillsBeds):
    pass


@dataclass
class BedwarsLate(WinsFinalsKillsBeds):
    single: BedwarsLateSingle = None
    double: BedwarsLateDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.finals = self.single.finals + self.double.finals
        self.kills = self.single.kills + self.double.kills
        self.beds = self.single.beds + self.double.beds



from datetime import datetime
from dataclasses import dataclass, field, InitVar
from typing import List, Dict

from ..constants import *
from .games import *
from .achievements import *

from .utils import (
    calculate_total_stats,
    calculate_total_xp,
    calculate_required_xp,
    alias_convert,
    GAME_ALIASES,
    GAME_STATS,
)

__all__ = [
    "MinecraftPlayer",
    "VoxylPlayer",
    "VoxylPlayerInfo",
    "VoxylPlayerOverall",
    "VoxylPlayerGames",
    "VoxylPlayerGuild",
    "VoxylPlayerAchievements",
    "StatsLeaderboardPlayer",
    "LevelLeaderboardPlayer",
    "WeightedWinsLeaderboardPlayer",
    "TechniqueLeaderboardPlayer",
    "PeriodicLeaderboardPlayer",
    "GuildMember",
    "GuildMemberWithID",
]

games = {
    "bedRush": BedRush,
    "bedwalls": Bedwalls,
    "bedwarsLate": BedwarsLate,
    "bedwarsMega": BedwarsMega,
    "bedwarsNormal": BedwarsNormal,
    "bedwarsRush": BedwarsRush,
    "bowFight": BowFight,
    "bridgeFight": BridgeFight,
    "flatFight": FlatFight,
    "fourWayBridge": FourWayBridge,
    "groundFight": GroundFight,
    "ladderFight": LadderFight,
    "miniwars": Miniwars,
    "obstacles": Obstacles,
    "partyGames": PartyGames,
    "pearlFight": PearlFight,
    "rankedFoursPractice": RankedFoursPractice,
    "resourceCollect": ResourceCollect,
    "resourceOldCollect": ResourceOldCollect,
    "stickFight": StickFight,
    "sumo": Sumo,
    "sumoDuels": SumoDuels,
    "voidFight": VoidFight,
}


@dataclass
class MinecraftPlayer:
    _data: InitVar[Dict] = None
    name: str = None
    uuid: str = None
    formattedUUID: str = None

    def __post_init__(self, _data: dict):
        if not _data:
            return

        self.name = _data.get("username", None)
        self.formattedUUID = _data.get("uuid", None)
        self.uuid = self.formattedUUID.replace("-", "") if self.formattedUUID else None


@dataclass
class GuildMember:
    _data: InitVar[Dict] = None
    uuid: str = None
    role: str = None
    join_date_raw: int = 0
    join_date: datetime = 0

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "GUILD_MEMBER")
        for i in data:
            setattr(self, i, data[i])

        self.join_date = datetime.fromtimestamp(self.join_date_raw).strftime(
            "%I:%M %p on %B %d, %Y"
        )


@dataclass
class GuildMemberWithID(GuildMember):
    guild_id: int = 0


@dataclass
class VoxylPlayer:
    _gen_data: InitVar[Dict] = None
    _over_data: InitVar[Dict] = None
    _game_data: InitVar[Dict] = None
    _guild_data: InitVar[Dict] = None
    _ach_data: InitVar[List] = None
    uuid: str = None
    name: str = None
    last_login_time_raw: int = 0
    last_login_time: datetime = None
    rank: str = None
    level: int = 0
    xp: int = 0
    required_xp: int = 0
    total_xp: int = 0
    weightedwins: int = 0
    wins: int = 0
    finals: int = 0
    kills: int = 0
    beds: int = 0
    guild: GuildMemberWithID = None

    bedRush: BedRush = None
    bedwalls: Bedwalls = None
    bedwarsLate: BedwarsLate = None
    bedwarsMega: BedwarsMega = None
    bedwarsNormal: BedwarsNormal = None
    bedwarsRush: BedwarsRush = None
    bowFight: BowFight = None
    bridgeFight: BridgeFight = None
    flatFight: FlatFight = None
    fourWayBridge: FourWayBridge = None
    groundFight: GroundFight = None
    ladderFight: LadderFight = None
    miniwars: Miniwars = None
    obstacles: Obstacles = None
    partyGames: PartyGames = None
    pearlFight: PearlFight = None
    rankedFoursPractice: RankedFoursPractice = None
    resourceCollect: ResourceCollect = None
    resourceOldCollect: ResourceOldCollect = None
    stickFight: StickFight = None
    sumo: Sumo = None
    sumoDuels: SumoDuels = None
    voidFight: VoidFight = None

    achievements: List[Achievement] = field(default_factory=list)

    def __post_init__(
        self,
        _gen_data: dict,
        _over_data: dict,
        _game_data: dict,
        _guild_data: dict,
        _ach_data: List,
    ):
        if not _gen_data:
            self.guild = GuildMemberWithID(**self.guild) if self.guild else None

            game_data = {game: self.__dict__[game] for game in games}

            for game, model in games.items():
                data = {}
                for i in model.__dataclass_fields__:
                    if "voxyl" not in str(model.__dataclass_fields__[i].type):
                        data[i] = game_data[game][i]
                        continue

                    data[i] = model.__dataclass_fields__[i].type(**game_data[game][i])

                setattr(self, game, model(**data))

            return

        data = alias_convert({**_gen_data, **_over_data}, "PLAYER")
        for i in data:
            setattr(self, i, data[i])

        total = calculate_total_stats(_game_data)
        self.wins = total.get("wins", 0)
        self.finals = total.get("finals", 0)
        self.kills = total.get("kills", 0)
        self.beds = total.get("beds", 0)

        self.required_xp = calculate_required_xp(self.level)
        self.total_xp = calculate_total_xp(self.level, self.xp)

        self.last_login_time = datetime.fromtimestamp(
            self.last_login_time_raw
        ).strftime("%I:%M %p on %B %d, %Y")

        for game, model in games.items():
            data = alias_convert(_game_data, game, model)
            setattr(self, game, model(**data))

        self.achievements = _ach_data

        self.guild = (
            GuildMemberWithID(uuid=self.uuid, _data=_guild_data)
            if _guild_data
            else None
        )


@dataclass
class VoxylPlayerInfo:
    _data: InitVar[Dict] = None
    uuid: str = None
    name: str = None
    last_login_time_raw: int = 0
    last_login_time: datetime = None
    rank: str = None

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "PLAYER")
        for i in data:
            setattr(self, i, data[i])

        self.last_login_time = datetime.fromtimestamp(
            self.last_login_time_raw
        ).strftime("%I:%M %p on %B %d, %Y")


@dataclass
class VoxylPlayerOverall:
    _data: InitVar[Dict] = None
    uuid: str = None
    name: str = None
    level: int = 0
    xp: int = 0
    required_xp: int = 0
    total_xp: int = 0

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "PLAYER")
        for i in data:
            setattr(self, i, data[i])

        self.required_xp = calculate_required_xp(self.level)
        self.total_xp = calculate_total_xp(self.level, self.xp)


@dataclass
class VoxylPlayerGames:
    _data: InitVar[Dict] = None
    uuid: str = None
    name: str = None
    wins: int = 0
    finals: int = 0
    kills: int = 0
    beds: int = 0
    bedRush: BedRush = None
    bedwalls: Bedwalls = None
    bedwarsLate: BedwarsLate = None
    bedwarsMega: BedwarsMega = None
    bedwarsNormal: BedwarsNormal = None
    bedwarsRush: BedwarsRush = None
    bowFight: BowFight = None
    bridgeFight: BridgeFight = None
    flatFight: FlatFight = None
    fourWayBridge: FourWayBridge = None
    groundFight: GroundFight = None
    ladderFight: LadderFight = None
    miniwars: Miniwars = None
    obstacles: Obstacles = None
    partyGames: PartyGames = None
    pearlFight: PearlFight = None
    rankedFoursPractice: RankedFoursPractice = None
    resourceCollect: ResourceCollect = None
    resourceOldCollect: ResourceOldCollect = None
    stickFight: StickFight = None
    sumo: Sumo = None
    sumoDuels: SumoDuels = None
    voidFight: VoidFight = None

    def __post_init__(self, _data: dict):
        if not _data:
            return

        total = calculate_total_stats(_data)
        self.wins = total.get("wins", 0)
        self.finals = total.get("finals", 0)
        self.kills = total.get("kills", 0)
        self.beds = total.get("beds", 0)

        for game, model in games.items():
            data = alias_convert(_data, game, model)
            setattr(self, game, model(**data))


@dataclass
class VoxylPlayerGuild:
    _data: InitVar[Dict] = None
    uuid: str = None
    name: str = None
    guild_id: int = 0
    role: str = None
    join_date_raw: int = 0
    join_date: datetime = None

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "PLAYER")
        for i in data:
            setattr(self, i, data[i])

        self.join_date = datetime.fromtimestamp(self.join_date_raw).strftime(
            "%I:%M %p on %B %d, %Y"
        )


@dataclass
class VoxylPlayerAchievements:
    _data: InitVar[Dict] = None
    uuid: str = None
    name: str = None
    achievements: List[Achievement] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        self.achievements = _data


@dataclass
class StatsLeaderboardPlayer:
    _data: InitVar[Dict] = None
    uuid: str = None
    level: int = 0
    weightedwins: int = 0
    position: int = 0

    def __post_init__(self, _data: dict):
        data = alias_convert(_data, "LEADERBOARD")
        for i in data:
            setattr(self, i, data[i])


@dataclass
class LevelLeaderboardPlayer(StatsLeaderboardPlayer):
    leaderboard: str = "level"


@dataclass
class WeightedWinsLeaderboardPlayer(StatsLeaderboardPlayer):
    leaderboard: str = "weightedwins"


@dataclass
class TechniqueLeaderboardPlayer:
    _data: InitVar[Dict] = None
    uuid: str = None
    technique_time: float = 0
    date_submitted_raw: int = 0
    date_submitted: datetime = 0
    position: int = 0

    def __post_init__(self, _data: dict):
        data = alias_convert(_data, "LEADERBOARD")
        for i in data:
            setattr(self, i, data[i])

        self.technique_time = int(self.technique_time)
        self.date_submitted = datetime.fromtimestamp(self.date_submitted_raw).strftime(
            "%I:%M %p on %B %d, %Y"
        )


@dataclass
class PeriodicLeaderboardPlayer:
    _data: InitVar[Dict] = None
    uuid: str = None
    wins: int = 0
    position: int = 0

    def __post_init__(self, _data: dict):
        data = alias_convert(_data, "LEADERBOARD")
        for i in data:
            setattr(self, i, data[i])



from datetime import datetime
from dataclasses import dataclass, field, InitVar
from typing import Dict, List

from ..constants import *
from .utils import alias_convert
from .player import GuildMember
from ..errors import GuildMemberNotFound

__all__ = ["Guild", "GuildInfo", "GuildMembers", "LeaderboardGuild"]


@dataclass
class Guild:
    _info_data: InitVar[Dict] = None
    _member_data: InitVar[Dict] = None
    guild_id: int = 0
    name: str = None
    description: str = None
    gxp: int = 0
    member_count: int = 0
    org_owner: str = None
    creation_date: datetime = 0
    creation_date_raw: int = 0
    all_members: List[GuildMember] = field(default_factory=list)
    owner: GuildMember = None
    admins: List[GuildMember] = field(default_factory=list)
    moderators: List[GuildMember] = field(default_factory=list)
    members: List[GuildMember] = field(default_factory=list)

    def __post_init__(self, _info_data: dict, _member_data: dict):
        if not _info_data:
            return

        data = alias_convert(_info_data, "GUILD")
        for i in data:
            setattr(self, i, data[i])

        self.creation_date = datetime.fromtimestamp(self.creation_date_raw).strftime(
            "%I:%M %p on %B %d, %Y"
        )

        for i in _member_data:
            self.all_members.append(GuildMember(i))

            if i["role"] == "OWNER":
                self.owner = GuildMember(i)

            else:
                self.__dict__[f'{i["role"].lower()}s'].append(GuildMember(i))

    def get_guild_member(self, uuid: str):
        for i in self.all_members:
            if i.uuid.replace("-", "") == uuid.replace("-", ""):
                return i

        raise GuildMemberNotFound(uuid, self.name)


@dataclass
class GuildInfo:
    _data: InitVar[Dict] = None
    guild_id: int = 0
    name: str = None
    description: str = None
    gxp: int = 0
    member_count: int = 0
    org_owner: str = None
    creation_date: datetime = 0
    creation_date_raw: int = 0

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "GUILD")
        for i in data:
            setattr(self, i, data[i])

        self.creation_date = datetime.fromtimestamp(self.creation_date_raw).strftime(
            "%I:%M %p on %B %d, %Y"
        )


@dataclass
class GuildMembers:
    _data: InitVar[Dict] = None
    all_members: List[GuildMember] = field(default_factory=list)
    owner: GuildMember = None
    admins: List[GuildMember] = field(default_factory=list)
    moderators: List[GuildMember] = field(default_factory=list)
    members: List[GuildMember] = field(default_factory=list)

    def __post_init__(self, _data: dict):
        if not _data:
            return

        for i in _data:
            self.all_members.append(GuildMember(i))

        for i in _data:
            if i["role"] == "OWNER":
                self.owner = GuildMember(i)

            else:
                self.__dict__[f'{i["role"].lower()}s'].append(GuildMember(i))

    def get_guild_member(self, uuid: str):
        for i in self.all_members:
            if i.uuid.replace("-", "") == uuid.replace("-", ""):
                return i

        raise GuildMemberNotFound(uuid)


@dataclass
class LeaderboardGuild:
    _data: InitVar[Dict] = None
    guild_id: int = 0
    tag: str = None
    gxp: int = 0
    name: str = None
    position: int = 0

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "GUILD")
        for i in data:
            setattr(self, i, data[i])



from typing import Union


class VoxylException(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class APIError(VoxylException):
    def __init__(self, resp: dict, api: str, message: str = None):
        if not message:
            message = f"An unknown error has occurred in the {api} API"

        self.resp = resp
        self.api = api
        super().__init__(message)


class ArgumentError(VoxylException):
    def __init__(self, message: str):
        super().__init__(message)


class ClosedSession(VoxylException):
    def __init__(self, message=None):
        if not message:
            message = "Session closed"

        super().__init__(message)


class GuildMemberNotFound(ArgumentError):
    def __init__(self, member: str, guild_name: str = None):
        super().__init__(
            f"'{member}' was not found in the guild '{guild_name}'"
            if guild_name
            else f"'{member}' was not found in the guild"
        )


class GuildNotFound(APIError):
    def __init__(self, api: str, guild: str, message: str = None):
        super().__init__(
            resp=None,
            api=api,
            message=message if message else f"Guild not found: '{guild}'",
        )


class InvalidAPIKey(ArgumentError):
    def __init__(self, api_key: str, message: str = None):
        if not message:
            message = f"Invalid API Key: '{api_key}'"

        super().__init__(message)


class InvalidGameReference(ArgumentError):
    def __init__(self, game: str, message: str = None):
        super().__init__(message if message else f"Invalid game: '{game}'")


class InvalidLeaderboardNumber(ArgumentError):
    def __init__(self, num: Union[int, str], message: str = None):
        super().__init__(
            message
            if message
            else f"Invalid number for leaderboard player amount: '{num}'"
        )


class InvalidTechnique(ArgumentError):
    def __init__(self, technique: str, message: str = None):
        super().__init__(message if message else f"Invalid technique: '{technique}'")


class KeyNotFound(ArgumentError):
    def __init__(self, key: str):
        super().__init__(f"The key, '{key}', was not found")


class KeyRequired(InvalidAPIKey):
    def __init__(self, path=None):
        super().__init__(f"{path} requires an API Key")


class PlayerNotFound(APIError):
    def __init__(self, api: str, player: str, message: str = None):
        super().__init__(
            resp=None,
            api=api,
            message=message if message else f"Player not found: '{player}'",
        )


class RateLimitError(APIError):
    def __init__(self, api: str, resp: dict):
        super().__init__(
            resp=resp, api=api, message=f"You are being rate limited ({api})"
        )


class TimeoutError(APIError):
    def __init__(self, api: str):
        super().__init__(resp=None, api=api, message=f"API Timeout {api}")



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["StickFight", "StickFightSingle", "StickFightDouble"]


@dataclass
class StickFightSingle(WinsKills):
    pass


@dataclass
class StickFightDouble(WinsKills):
    pass


@dataclass
class StickFight(WinsKills):
    single: StickFightSingle = None
    double: StickFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsFinalsKillsBeds

__all__ = ["GroundFight", "GroundFightSingle", "GroundFightDouble"]


@dataclass
class GroundFightSingle(WinsFinalsKillsBeds):
    pass


@dataclass
class GroundFightDouble(WinsFinalsKillsBeds):
    pass


@dataclass
class GroundFight(WinsFinalsKillsBeds):
    single: GroundFightSingle = None
    double: GroundFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.finals = self.single.finals + self.double.finals
        self.kills = self.single.kills + self.double.kills
        self.beds = self.single.beds + self.double.beds



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["ResourceCollect", "ResourceCollectSingle", "ResourceCollectDouble"]


@dataclass
class ResourceCollectSingle(WinsKills):
    pass


@dataclass
class ResourceCollectDouble(WinsKills):
    pass


@dataclass
class ResourceCollect(WinsKills):
    single: ResourceCollectSingle = None
    double: ResourceCollectDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["SumoDuels", "SumoDuelsSingle", "SumoDuelsDouble"]


@dataclass
class SumoDuelsSingle(WinsKills):
    pass


@dataclass
class SumoDuelsDouble(WinsKills):
    pass


@dataclass
class SumoDuels(WinsKills):
    single: SumoDuelsSingle = None
    double: SumoDuelsDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsFinalsKillsBeds

__all__ = ["BridgeFight", "BridgeFightSingle", "BridgeFightDouble", "BridgeFightComp"]


@dataclass
class BridgeFightSingle(WinsFinalsKillsBeds):
    pass


@dataclass
class BridgeFightDouble(WinsFinalsKillsBeds):
    pass


@dataclass
class BridgeFightComp(WinsFinalsKillsBeds):
    pass


@dataclass
class BridgeFight(WinsFinalsKillsBeds):
    single: BridgeFightSingle = None
    double: BridgeFightDouble = None
    comp: BridgeFightComp = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins + self.comp.wins
        self.finals = self.single.finals + self.double.finals + self.comp.finals
        self.kills = self.single.kills + self.double.kills + self.comp.kills
        self.beds = self.single.beds + self.double.beds + self.comp.beds



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["BowFight", "BowFightSingle", "BowFightDouble"]


@dataclass
class BowFightSingle(WinsKills):
    pass


@dataclass
class BowFightDouble(WinsKills):
    pass


@dataclass
class BowFight(WinsKills):
    single: BowFightSingle = None
    double: BowFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["BedRush", "BedRushSingle", "BedRushDouble"]


@dataclass
class BedRushSingle(WinsKills):
    pass


@dataclass
class BedRushDouble(WinsKills):
    pass


@dataclass
class BedRush(WinsKills):
    single: BedRushSingle = None
    double: BedRushDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["FlatFight", "FlatFightSingle", "FlatFightDouble"]


@dataclass
class FlatFightSingle(WinsKills):
    pass


@dataclass
class FlatFightDouble(WinsKills):
    pass


@dataclass
class FlatFight(WinsKills):
    single: FlatFightSingle = None
    double: FlatFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



import asyncio, aiohttp
from typing import Dict, List, Union, TypedDict

from voxyl.models import *

from .constants import *

from .errors import *


class Client:
    def __init__(self, keys: Union[str, List[str]] = None, loop=None, **options):
        self.loop = asyncio.get_event_loop() if loop is None else loop
        if isinstance(keys, str):
            self._keys = [keys]

        elif isinstance(keys, list):
            self._keys = keys

        else:
            raise InvalidAPIKey("Please provide a string or list of api keys")

        self._itr = iter(self._keys)

        self.timeout = options.get("timeout", 10)

        self._session = aiohttp.ClientSession(
            loop=self.loop, timeout=aiohttp.ClientTimeout(total=self.timeout)
        )

    async def __aenter__(self):
        if self._session.closed:
            self._session = aiohttp.ClientSession(loop=self.loop)

        return self

    async def __aexit__(self, *args):
        await self._session.close()

    def _next_key(self) -> str:
        if not self._keys:
            raise KeyRequired("method '_next_key'")

        try:
            return next(self._itr)

        except StopIteration:
            self._itr = iter(self._keys)
            return next(self._itr)

    async def _get(
        self,
        path: str,
        query_params: dict = None,
        path_params: list = None,
        key_required: bool = True,
    ) -> Dict:
        if self._session.closed:
            raise ClosedSession()

        if not query_params:
            query_params = {}

        if not path_params:
            path_params = []

        if key_required:
            if not self._keys:
                raise KeyRequired(path)

            query_params["api"] = self._next_key()

        try:
            response = await self._session.get(
                url=VOXYL_URLS[path] + "/".join(path_params), params=query_params
            )

        except:
            raise TimeoutError("Voxyl API")

        if response.status == 429:
            raise RateLimitError(path, response)

        elif response.status == 200:
            return await response.json(content_type=None)

        elif response.status == 400 or response.status == 403:
            text = await response.json(content_type=None)
            if "api" not in text["reason"].lower():
                raise APIError(resp=text, api=path, message=text["reason"])

            elif key_required:
                if query_params.get("api", None):
                    raise InvalidAPIKey(query_params.get("api"))

                else:
                    raise KeyRequired(path)

            else:
                raise APIError(
                    response,
                    "Voxyl API",
                    f"An unexpected error has occurred with the Voxyl API",
                )

        else:
            try:
                text = (await response.json(content_type=None)).get("reason")

            except:
                raise APIError(response, "Voxyl API")

            else:
                raise APIError(
                    response,
                    "Voxyl API",
                    f"An unexpected error has occurred with the Voxyl API: {text}",
                )

    async def close(self):
        await self._session.close()

    @property
    def keys(self) -> List[str]:
        return self._keys

    def add_key(self, key):
        if isinstance(key, str):
            self._keys.append(key)
            self._itr = iter(self._keys)

        else:
            raise KeyNotFound(key)

    def remove_key(self, key):
        if isinstance(key, str):
            try:
                self._keys.append(key)
                self._itr = iter(self._keys)

            except ValueError:
                pass

        else:
            raise InvalidAPIKey(key)

    def format_uuid(self, uuid: str):
        uuid = self.shorten_uuid(uuid)
        return (
            uuid[:8]
            + "-"
            + uuid[8:12]
            + "-"
            + uuid[12:16]
            + "-"
            + uuid[16:20]
            + "-"
            + uuid[20:]
        )

    def shorten_uuid(self, uuid: str):
        return uuid.replace("-", "")

    def calculate_required_xp(self, level: int):
        return calculate_required_xp(level)

    def calculate_total_xp(self, level: int, xp: int):
        return calculate_total_xp(level, xp / self.calculate_required_xp(level))

    async def _retrieve_minecraft_player(self, player: str) -> MinecraftPlayer:
        try:
            mojdata = await self._get(
                "minecraft", path_params=[player], key_required=False
            )

        except APIError:
            raise PlayerNotFound("minecraft", player)

        return MinecraftPlayer(_data=mojdata)

    async def _retrieve_voxyl_player_data(
        self, path: str, player: MinecraftPlayer
    ) -> dict:
        try:
            return await self._get(path, path_params=[player.formattedUUID])

        except APIError:
            raise PlayerNotFound(
                path,
                player.formattedUUID,
                f"Could not find voxyl player data on '{player.name}'",
            )

    async def _retrieve_historic_leaderboard(
        self, game: str, stat_type: str, period: str
    ):
        try:
            lb_info = await self._get(
                "leaderboard_game",
                path_params=[game],
                query_params={"type": stat_type, "period": period},
            )

        except APIError:
            raise InvalidGameReference(game)

        return PeriodicLeaderboard(
            _data=lb_info["players"], game=game, type=stat_type, period=period
        )

    async def get_minecraft_player(
        self, player: Union[str, MinecraftPlayer]
    ) -> MinecraftPlayer:
        if isinstance(player, str):
            try:
                return await self._retrieve_minecraft_player(player)

            except APIError:
                raise PlayerNotFound("minecraft", player)

        elif isinstance(player, MinecraftPlayer):
            return player

    async def get_voxyl_player(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayer:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            info = await self._retrieve_voxyl_player_data(
                "player_info", minecraft_player
            )
            over = await self._retrieve_voxyl_player_data(
                "player_overall", minecraft_player
            )
            game = await self._retrieve_voxyl_player_data(
                "player_game", minecraft_player
            )

        except APIError:
            raise PlayerNotFound(
                "player",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player data on '{minecraft_player.name}'",
            )

        try:
            guild = await self._retrieve_voxyl_player_data(
                "player_guild", minecraft_player
            )

        except APIError:
            guild = None

        try:
            achievements = (
                await self._retrieve_voxyl_player_data(
                    "player_achievements", minecraft_player
                )
            )["achievements"]
            all_achievements = await self.get_achievements()

            for i in range(len(achievements)):
                achievements[i] = next(
                    (x for x in all_achievements if x.achievement_id == achievements[i])
                )

        except APIError:
            achievements = []

        return VoxylPlayer(
            uuid=minecraft_player.formattedUUID,
            _gen_data=info,
            _over_data=over,
            _game_data=game["stats"],
            _guild_data=guild,
            _ach_data=achievements,
        )

    async def get_voxyl_player_info(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayerInfo:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            info = await self._retrieve_voxyl_player_data(
                "player_info", minecraft_player
            )

        except APIError:
            raise PlayerNotFound(
                "player_info",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player info data on '{minecraft_player.name}'",
            )

        return VoxylPlayerInfo(uuid=minecraft_player.formattedUUID, _data=info)

    async def get_voxyl_player_overall(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayerOverall:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            over = await self._retrieve_voxyl_player_data(
                "player_overall", minecraft_player
            )

        except APIError:
            raise PlayerNotFound(
                "player_overall",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player overall data on '{minecraft_player.name}'",
            )

        return VoxylPlayerOverall(
            uuid=minecraft_player.formattedUUID, name=minecraft_player.name, _data=over
        )

    async def get_voxyl_player_games(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayerGames:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            game = await self._retrieve_voxyl_player_data(
                "player_game", minecraft_player
            )

        except APIError:
            raise PlayerNotFound(
                "player_game",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player game data on '{minecraft_player.name}'",
            )

        return VoxylPlayerGames(
            uuid=minecraft_player.formattedUUID,
            name=minecraft_player.name,
            _data=game["stats"],
        )

    async def get_voxyl_player_guild(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayerGuild:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            guild = await self._retrieve_voxyl_player_data(
                "player_guild", minecraft_player
            )

        except APIError:
            raise PlayerNotFound(
                "player_guild",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player guild data '{minecraft_player.name}'",
            )

        return VoxylPlayerGuild(
            uuid=minecraft_player.formattedUUID, name=minecraft_player.name, _data=guild
        )

    async def get_voxyl_player_achievements(
        self, player: Union[str, MinecraftPlayer]
    ) -> VoxylPlayerAchievements:
        minecraft_player = await self.get_minecraft_player(player)

        try:
            achievements = (
                await self._retrieve_voxyl_player_data(
                    "player_achievements", minecraft_player
                )
            )["achievements"]
            all_achievements = await self.get_achievements()

            for i in range(len(achievements)):
                achievements[i] = next(
                    (x for x in all_achievements if x.achievement_id == achievements[i])
                )

        except APIError:
            raise PlayerNotFound(
                "player_achievements",
                minecraft_player.formattedUUID,
                f"Could not find voxyl player achievement data on '{minecraft_player.name}'",
            )

        return VoxylPlayerAchievements(
            uuid=minecraft_player.formattedUUID,
            name=minecraft_player.name,
            _data=achievements,
        )

    async def get_guild(self, identifier: str) -> Guild:
        try:
            guild_info = await self._get("guild_info", path_params=[identifier])
            guild_members = await self._get("guild_members", path_params=[identifier])

        except APIError:
            raise GuildNotFound(
                "guild",
                identifier,
                (
                    f"No guild found with id '{identifier[1:]}'"
                    if identifier[1:].isdigit()
                    else f"No guild found with tag '{identifier}'"
                ),
            )

        return Guild(_info_data=guild_info, _member_data=guild_members["members"])

    async def get_guild_info(self, identifier: str) -> GuildInfo:
        try:
            guild_info = await self._get("guild_info", path_params=[identifier])

        except APIError:
            raise GuildNotFound(
                "guild_info",
                identifier,
                (
                    f"No guild found with id '{identifier[1:]}'"
                    if identifier[1:].isdigit()
                    else f"No guild found with tag '{identifier}'"
                ),
            )

        return GuildInfo(_data=guild_info)

    async def get_guild_members(self, identifier: str) -> GuildMembers:
        try:
            guild_members = await self._get("guild_members", path_params=[identifier])

        except APIError:
            raise GuildNotFound(
                "guild_members",
                identifier,
                (
                    f"No guild found with id '{identifier[1:]}'"
                    if identifier[1:].isdigit()
                    else f"No guild found with tag '{identifier}'"
                ),
            )

        return GuildMembers(_data=guild_members["members"])

    async def get_guild_from_player(self, player: Union[str, MinecraftPlayer]) -> Guild:
        player = await self.get_voxyl_player_guild(player)

        try:
            return await self.get_guild(identifier=f"-{player.guild_id}")

        except APIError:
            raise GuildNotFound(
                "guild_player",
                f"-{player.guild_id}",
                f"No guild found with id '{player.guild_id}'",
            )

    async def get_guild_info_from_player(
        self, player: Union[str, MinecraftPlayer]
    ) -> GuildInfo:
        player = await self.get_voxyl_player_guild(player)

        try:
            return await self.get_guild_info(identifier=f"-{player.guild_id}")

        except APIError:
            raise GuildNotFound(
                "guild_info_player",
                f"-{player.guild_id}",
                f"No guild found with id '{player.guild_id}'",
            )

    async def get_guild_members_from_player(
        self, player: Union[str, MinecraftPlayer]
    ) -> GuildMembers:
        player = await self.get_voxyl_player_guild(player)

        try:
            return await self.get_guild_members(identifier=f"-{player.guild_id}")

        except APIError:
            raise GuildNotFound(
                "guild_members_player",
                f"-{player.guild_id}",
                f"No guild found with id '{player.guild_id}'",
            )

    async def get_level_leaderboard(self, num: int = 10) -> LevelLeaderboard:
        try:
            lb_info = await self._get(
                "leaderboard_level", query_params={"type": "level", "num": num}
            )

        except APIError:
            raise InvalidLeaderboardNumber(num)

        return LevelLeaderboard(_data=lb_info["players"])

    async def get_weightedwins_leaderboard(
        self, num: int = 10
    ) -> WeightedWinsLeaderboard:
        try:
            lb_info = await self._get(
                "leaderboard_weightedwins",
                query_params={"type": "weightedwins", "num": num},
            )

        except APIError:
            raise InvalidLeaderboardNumber(num)

        return WeightedWinsLeaderboard(_data=lb_info["players"])

    async def get_guild_leaderboard(self, num: int = 10) -> GuildLeaderboard:
        try:
            lb_info = await self._get("leaderboard_guild", query_params={"num": num})

        except APIError:
            raise InvalidLeaderboardNumber(num)

        return GuildLeaderboard(_data=lb_info["guilds"])

    async def get_technique_leaderboard(
        self, technique: str = "safetower"
    ) -> TechniqueLeaderboard:
        try:
            lb_info = await self._get(
                "leaderboard_technique", query_params={"technique": technique}
            )

        except APIError:
            raise InvalidTechnique(technique)

        return TechniqueLeaderboard(_data=lb_info["guilds"])

    async def get_daily_wins_leaderboard(self, game: str) -> PeriodicLeaderboard:
        return await self._retrieve_historic_leaderboard(
            game=game, stat_type="wins", period="daily"
        )

    async def get_daily_winstreaks_leaderboard(self, game: str) -> PeriodicLeaderboard:
        return await self._retrieve_historic_leaderboard(
            game=game, stat_type="winstreaks", period="daily"
        )

    async def get_weekly_wins_leaderboard(self, game: str) -> PeriodicLeaderboard:
        return await self._retrieve_historic_leaderboard(
            game=game, stat_type="wins", period="weekly"
        )

    async def get_weekly_winstreaks_leaderboard(self, game: str) -> PeriodicLeaderboard:
        return await self._retrieve_historic_leaderboard(
            game=game, stat_type="winstreaks", period="weekly"
        )

    async def get_achievements(
        self, id: int = None, name: str = None, description: str = None, xp: int = None
    ) -> List[Achievement]:
        achievements = (await self._get("achievements"))["info"]

        if (
            id is not None
            or name is not None
            or description is not None
            or xp is not None
        ):
            achievements = [
                ach
                for ach in achievements
                if (
                    id == ach["id"]
                    or (name and name.lower() in ach["name"].lower())
                    or (description and description.lower() in ach["desc"].lower())
                    or xp == ach["xp"]
                )
            ]

        for i in range(len(achievements)):
            achievements[i] = Achievement(achievements[i])

        return achievements

    async def get_announcements(self) -> dict:
        return await self._get("announcements")

    def get_techniques(self, technique: str = None) -> List[str]:
        return (
            VOXYL_TECHNIQUES.index(technique.lower()) if technique else VOXYL_TECHNIQUES
        )

    def get_games(
        self, game: str = None
    ) -> Dict[str, Dict[Dict[str, str], Dict[str, float]]]:
        return (
            VOXYL_GAMES.get(VOXYL_GAME_NAMES_LOWER.get(game.lower(), None), None)
            if game
            else VOXYL_GAMES
        )

    def get_ranks(
        self, rank: str = None
    ) -> Dict[str, Dict[str, Union[str, List[int]]]]:
        return VOXYL_RANKS.get(rank.lower(), None) if rank else VOXYL_RANKS

    def get_stars(
        self, prestige: str = None
    ) -> Dict[str, Dict[str, Union[str, int, List[int]]]]:
        return (
            VOXYL_PRESTIGES.get(prestige.lower(), None) if prestige else VOXYL_PRESTIGES
        )



from dataclasses import dataclass

__all__ = ["WinsFinalsKillsBeds", "WinsKills", "Wins"]


@dataclass
class WinsFinalsKillsBeds:
    wins: int = 0
    finals: int = 0
    kills: int = 0
    beds: int = 0


@dataclass
class WinsKills:
    wins: int = 0
    kills: int = 0


@dataclass
class Wins:
    wins: int = 0



VOXYL_BASE = "https://api.voxyl.net"
VOXYL_URLS = {
    "minecraft": "https://api.ashcon.app/mojang/v2/user/",
    "player_info": VOXYL_BASE + "/player/info/",
    "player_overall": VOXYL_BASE + "/player/stats/overall/",
    "player_game": VOXYL_BASE + "/player/stats/game/",
    "player_guild": VOXYL_BASE + "/player/guild/",
    "player_achievements": VOXYL_BASE + "/achievements/player/",
    "guild_info": VOXYL_BASE + "/guild/info/",
    "guild_members": VOXYL_BASE + "/guild/members/",
    "leaderboard_guild": VOXYL_BASE + "/guild/top/",
    "leaderboard_level": VOXYL_BASE + "/leaderboard/normal/",
    "leaderboard_weightedwins": VOXYL_BASE + "/leaderboard/normal/",
    "leaderboard_technique": VOXYL_BASE + "/leaderboard/technique/",
    "leaderboard_game": VOXYL_BASE + "/leaderboard/game/",
    "announcements": VOXYL_BASE + "/announcement/all/",
    "achievements": VOXYL_BASE + "/achievements/info/",
}

VOXYL_TECHNIQUES = [
    "wallclutch",
    "highclutch",
    "sideclutch",
    "knockbackclutch",
    "jumparound",
    "safetower",
    "speedclutch",
    "ladderclutch",
    "hitclutch",
    "brokenwallrun",
    "bridgestart",
    "knockbackwallclutch",
]

VOXYL_GAMES = {
    "bridgesSingle": {
        "formatted_name": "1v1 Bed Bridge Fight",
        "max_players": 2,
        "weight": 0.55,
    },
    "bridgesDouble": {
        "formatted_name": "2v2 Bed Bridge Fight",
        "max_players": 4,
        "weight": 0.69,
    },
    "compBridgeSingle": {
        "formatted_name": "Competitive Bed Bridge Fight",
        "max_players": 2,
        "weight": 1.24,
    },
    "sumo": {"formatted_name": "Block Sumo", "max_players": 12, "weight": 3.93},
    "betaSumo": {"formatted_name": "Beta Block Sumo", "max_players": 16, "weight": 5.0},
    "sumoDuelsSolo": {
        "formatted_name": "1v1 Block Sumo Duels",
        "max_players": 2,
        "weight": 0.92,
    },
    "sumoDuelsDouble": {
        "formatted_name": "2v2 Block Sumo Duels",
        "max_players": 4,
        "weight": 1.26,
    },
    "obstacleSingle": {"formatted_name": "Obstacles", "max_players": 2, "weight": 1.2},
    "resourceSingle": {
        "formatted_name": "1v1 Resource Collect",
        "max_players": 2,
        "weight": 1.2,
    },
    "resourceDouble": {
        "formatted_name": "2v2 Resource Collect",
        "max_players": 4,
        "weight": 2.17,
    },
    "resourceOldSingle": {
        "formatted_name": "1v1 Old Resource Collect",
        "max_players": 2,
        "weight": 1.2,
    },
    "resourceOldDouble": {
        "formatted_name": "2v2 Old Resource Collect",
        "max_players": 4,
        "weight": 2.69,
    },
    "groundSingle": {
        "formatted_name": "1v1 Ground Fight",
        "max_players": 2,
        "weight": 0.46,
    },
    "groundDouble": {
        "formatted_name": "2v2 Ground Fight",
        "max_players": 4,
        "weight": 0.38,
    },
    "voidSingle": {
        "formatted_name": "1v1 Void Fight",
        "max_players": 2,
        "weight": 0.73,
    },
    "voidDouble": {
        "formatted_name": "2v2 Void Fight",
        "max_players": 4,
        "weight": 0.98,
    },
    "bedwarsNormalSingle": {
        "formatted_name": "1v1 Bedwars Normal",
        "max_players": 2,
        "weight": 1.0,
    },
    "bedwarsNormalDouble": {
        "formatted_name": "2v2 Bedwars Normal",
        "max_players": 4,
        "weight": 1.19,
    },
    "bedwarsMegaSolo": {
        "formatted_name": "Bedwars Mega Solo",
        "max_players": 8,
        "weight": 4.77,
    },
    "bedwarsMegaDouble": {
        "formatted_name": "Bedwars Mega Doubles",
        "max_players": 16,
        "weight": 4.84,
    },
    "bedwarsLateSingle": {
        "formatted_name": "1v1 Bedwars Late Game",
        "max_players": 2,
        "weight": 1.09,
    },
    "bedwarsLateDouble": {
        "formatted_name": "2v2 Bedwars Late Game",
        "max_players": 4,
        "weight": 1.26,
    },
    "bedwarsRushSolo": {
        "formatted_name": "1v1 Bedwars Rush Duels",
        "max_players": 2,
        "weight": 1.09,
    },
    "bedwarsRushDouble": {
        "formatted_name": "2v2 Bedwars Rush Duels",
        "max_players": 4,
        "weight": 1.26,
    },
    "stickFightSingle": {
        "formatted_name": "1v1 Stick Fight",
        "max_players": 2,
        "weight": 0.61,
    },
    "stickFightDouble": {
        "formatted_name": "2v2 Stick Fight",
        "max_players": 4,
        "weight": 0.66,
    },
    "flatFightSingle": {
        "formatted_name": "1v1 Flat Fight",
        "max_players": 2,
        "weight": 0.68,
    },
    "flatFightDouble": {
        "formatted_name": "2v2 Flat Fight",
        "max_players": 4,
        "weight": 0.88,
    },
    "ladderFightSingle": {
        "formatted_name": "1v1 Ladder Fight",
        "max_players": 2,
        "weight": 0.67,
    },
    "ladderFightDouble": {
        "formatted_name": "2v2 Ladder Fight",
        "max_players": 4,
        "weight": 0.8,
    },
    "partyGames": {"formatted_name": "Party Games", "max_players": 12, "weight": 7.92},
    "rankedFoursPractice": {
        "formatted_name": "4v4 Ranked Practice",
        "max_players": 8,
        "weight": 1.27,
    },
    "bowFightSingle": {
        "formatted_name": "1v1 Bow Fight",
        "max_players": 2,
        "weight": 0.55,
    },
    "bowFightDouble": {
        "formatted_name": "2v2 Bow Fight",
        "max_players": 4,
        "weight": 0.62,
    },
    "pearlFightSingle": {
        "formatted_name": "1v1 Pearl Fight",
        "max_players": 2,
        "weight": 0.5,
    },
    "pearlFightDouble": {
        "formatted_name": "2v2 Pearl Fight",
        "max_players": 4,
        "weight": 0.62,
    },
    "bedRushSingle": {
        "formatted_name": "1v1 Bed Rush",
        "max_players": 2,
        "weight": 0.75,
    },
    "bedRushDouble": {
        "formatted_name": "2v2 Bed Rush",
        "max_players": 4,
        "weight": 0.9,
    },
    "bedwalls": {"formatted_name": "Bedwalls", "max_players": 64, "weight": 20.0},
    "bedwarsEightSolo": {
        "formatted_name": "Bedwars Eight Solo",
        "max_players": 8,
        "weight": 1.21,
    },
    "miniwarsSolo": {
        "formatted_name": "Miniwars Solo",
        "max_players": 4,
        "weight": 2.79,
    },
    "miniwarsDouble": {
        "formatted_name": "Miniwars Double",
        "max_players": 8,
        "weight": 4.22,
    },
    "fourWayBridgeSingle": {
        "formatted_name": "Four Way Bridge Fight",
        "max_players": 4,
        "weight": 0.84,
    },
}

VOXYL_GAME_NAMES_LOWER = {
    "bedrushdouble": "bedRushDouble",
    "bedrushsingle": "bedRushSingle",
    "bedwalls": "bedwalls",
    "bedwarslatedouble": "bedwarsLateDouble",
    "bedwarslatesingle": "bedwarsLateSingle",
    "bedwarsmegadouble": "bedwarsMegaDouble",
    "bedwarsmegasolo": "bedwarsMegaSolo",
    "bedwarsnormaldouble": "bedwarsNormalDouble",
    "bedwarsnormalsingle": "bedwarsNormalSingle",
    "bedwarsrushdouble": "bedwarsRushDouble",
    "bedwarsrushsolo": "bedwarsRushSolo",
    "betasumo": "betaSumo",
    "bowfightdouble": "bowFightDouble",
    "bowfightsingle": "bowFightSingle",
    "bridgesdouble": "bridgesDouble",
    "bridgessingle": "bridgesSingle",
    "compbridgesingle": "compBridgeSingle",
    "flatfightdouble": "flatFightDouble",
    "flatfightsingle": "flatFightSingle",
    "fourwaybridgesingle": "fourWayBridgeSingle",
    "grounddouble": "groundDouble",
    "groundsingle": "groundSingle",
    "ladderfightdouble": "ladderFightDouble",
    "ladderfightsingle": "ladderFightSingle",
    "miniwarsdouble": "miniwarsDouble",
    "miniwarssolo": "miniwarsSolo",
    "obstaclesingle": "obstacleSingle",
    "partygames": "partyGames",
    "pearlfightdouble": "pearlFightDouble",
    "pearlfightsingle": "pearlFightSingle",
    "rankedfourspractice": "rankedFoursPractice",
    "resourcedouble": "resourceDouble",
    "resourcesingle": "resourceSingle",
    "resourceolddouble": "resourceOldDouble",
    "resourceoldsingle": "resourceOldSingle",
    "stickfightdouble": "stickFightDouble",
    "stickfightsingle": "stickFightSingle",
    "sumoduelsdouble": "sumoDuelsDouble",
    "sumoduelssolo": "sumoDuelsSolo",
    "sumo": "sumo",
    "voiddouble": "voidDouble",
    "voidsingle": "voidSingle",
}

VOXYL_RANKS = {
    "default": {
        "rgb": [85, 85, 85],
        "hex": "#555555",
    },
    "adept": {
        "rgb": [5, 172, 32],
        "hex": "#05ac20",
    },
    "expert": {
        "rgb": [7, 121, 209],
        "hex": "#0779d1",
    },
    "master": {
        "rgb": [247, 182, 15],
        "hex": "#f7b60f",
    },
    "youtube": {
        "rgb": [230, 3, 3],
        "hex": "#e60303",
    },
    "builder": {
        "rgb": [221, 101, 221],
        "hex": "#dd65dd",
    },
    "appeal": {
        "rgb": [245, 187, 187],
        "hex": "#f5bbbb",
    },
    "dev": {
        "rgb": [85, 255, 85],
        "hex": "#55ff55",
    },
    "helper": {
        "rgb": [90, 209, 90],
        "hex": "#5ad15a",
    },
    "trainee": {
        "rgb": [90, 209, 90],
        "hex": "#5ad15a",
    },
    "mod": {
        "rgb": [250, 241, 107],
        "hex": "#faf16b",
    },
    "srmod": {
        "rgb": [255, 163, 67],
        "hex": "#ffa343",
    },
    "manager": {
        "rgb": [220, 16, 16],
        "hex": "#dc1010",
    },
    "admin": {
        "rgb": [221, 64, 65],
        "hex": "#dd4041",
    },
    "owner": {"rgb": [252, 119, 126], "hex": "#fc777e"},
}

VOXYL_PRESTIGES = {
    "none": {"star": 0, "rgb": [48, 48, 48], "hex": "#303030"},
    "iron": {"star": 100, "rgb": [64, 64, 64], "hex": "#404040"},
    "gold": {"star": 200, "rgb": [62, 46, 4], "hex": "#3e2e04"},
    "diamond": {"star": 300, "rgb": [0, 64, 64], "hex": "#004040"},
    "emerald": {"star": 400, "rgb": [14, 44, 24], "hex": "#0e2c18"},
    "sapphire": {"star": 500, "rgb": [0, 43, 43], "hex": "#002b2b"},
    "ruby": {"star": 600, "rgb": [43, 0, 0], "hex": "#2b0000"},
    "crystal": {"star": 700, "rgb": [64, 21, 64], "hex": "#401540"},
    "opal": {"star": 800, "rgb": [21, 21, 64], "hex": "#151540"},
    "amethyst": {"star": 900, "rgb": [43, 0, 0], "hex": "#2b0000"},
}



"""Voxyl API Wrapper

A  simple python wrapper for the Voxyl API

"""

__title__ = "voxapiwrapper"
__author__ = "_lightninq"
__license__ = "MIT"
__copyright__ = "Copyright 2022 BWP-Stats"
__version__ = "1.0.0"

from .errors import *
from .client import Client



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["ResourceOldCollect", "ResourceOldCollectSingle", "ResourceOldCollectDouble"]


@dataclass
class ResourceOldCollectSingle(WinsKills):
    pass


@dataclass
class ResourceOldCollectDouble(WinsKills):
    pass


@dataclass
class ResourceOldCollect(WinsKills):
    single: ResourceOldCollectSingle = None
    double: ResourceOldCollectDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass, field, InitVar
from typing import Dict

from ..constants import *
from .utils import alias_convert

__all__ = ["Achievement"]


@dataclass
class Achievement:
    _data: InitVar[Dict] = None
    achievement_id: int = 0
    name: str = None
    description: str = None
    xp: int = 0

    def __post_init__(self, _data: dict):
        if not _data:
            return

        data = alias_convert(_data, "ACHIEVEMENT")
        for i in data:
            setattr(self, i, data[i])



from setuptools import setup, find_packages
import pathlib


HERE = pathlib.Path(__file__).parent
with open(f"{HERE}/README.md") as f:
    README = f.read()


setup(
    name="voxylstats",
    version="1.2.5",
    description="A simple python wrapper for the Voxyl API",
    long_description=README,
    author="_lightninq & firestarad",
    license="MIT",
    packages=["voxyl", "voxyl.models", "voxyl.models.games"],
)



from dataclasses import dataclass
from .default_game_structure import WinsKills

__all__ = ["PearlFight", "PearlFightSingle", "PearlFightDouble"]


@dataclass
class PearlFightSingle(WinsKills):
    pass


@dataclass
class PearlFightDouble(WinsKills):
    pass


@dataclass
class PearlFight(WinsKills):
    single: PearlFightSingle = None
    double: PearlFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.kills = self.single.kills + self.double.kills



from dataclasses import dataclass
from .default_game_structure import WinsFinalsKillsBeds

__all__ = ["LadderFight", "LadderFightSingle", "LadderFightDouble"]


@dataclass
class LadderFightSingle(WinsFinalsKillsBeds):
    pass


@dataclass
class LadderFightDouble(WinsFinalsKillsBeds):
    pass


@dataclass
class LadderFight(WinsFinalsKillsBeds):
    single: LadderFightSingle = None
    double: LadderFightDouble = None

    def __post_init__(self):
        self.wins = self.single.wins + self.double.wins
        self.finals = self.single.finals + self.double.finals
        self.kills = self.single.kills + self.double.kills
        self.beds = self.single.beds + self.double.beds
